from functools import lru_cache, partial
import os.path

import github

from . import __version__ as check_in_version
from .github_checks_requests import NewCheckRequest, UpdateCheckRequest, to_gh_query


cache_once = partial(lru_cache, maxsize=1)

DEFAULT_USER_AGENT = f'check-in/{check_in_version} (+https://pypi.org/p/check-in)'


class GithubClient:
    def __init__(self, app_id, installation_id, private_key_file, repo_slug=None, user_agent_prefix=None):
        self._gh_int = get_github_integration(app_id, private_key_file)
        self._gh_client = get_installation_client(self._gh_int, installation_id)
        self._check_runs_base_uri = f'/repos/{_repo_slug}/check-runs'
        self._repo_slug = repo_slug
        self.user_agent = user_agent_prefix

    def __getattr__(self, key):
        return getattr(self._gh_client, key)

    @property
    def user_agent(self):
        return self._user_agent

    @user_agent.setter
    def user_agent(self, user_agent_prefix):
        self._user_agent = DEFAULT_USER_AGENT
        if user_agent_prefix:
            self._user_agent = f'{user_agent_prefix} built with {self._user_agent}'

    def _get_check_caller(self):
        rp = self._gh_client.get_repo(self._repo_slug)
        check_headers={
            'Accept': 'application/vnd.github.antiope-preview+json',
            'User-Agent': self.user_agent,
        }
        return partial(
            rp._requester.requestJsonAndCheck,
            headers=check_headers,
        )

    def _get_check_creator(self):
        checker = self._get_check_caller()
        return partial(
            checker,
            url=self._check_runs_base_uri,
            verb='POST',
        )

    def _get_check_updater(self, check_run_id):
        checker = self._get_check_caller()
        return partial(
            checker,
            url=f'{self._check_runs_base_uri}/{check_run_id}',
            verb='PATCH',
        )

    def post_check(self, head_branch, head_sha, req):
        check_creator = self._get_check_creator()
        post_parameters = to_gh_query(NewCheckRequest(head_branch, head_sha, **req))
        headers, data = check_creator(input=post_parameters)
        return data

    def update_check(self, check_run_id, req):
        check_updater = self._get_check_updater(check_run_id)
        patch_parameters = to_gh_query(UpdateCheckRequest(**req))
        headers, data = check_updater(input=patch_parameters)
        return data


class GithubAPI:
    def __init__(self, app_id, installation_id, private_key_file, repo_slug, user_agent_prefix=None):
        self.app_id = app_id
        self.installation_id = installation_id
        self.private_key_file = os.path.expanduser(os.path.expandvars(private_key_file))
        self.repo_slug = repo_slug
        self.user_agent_prefix = user_agent_prefix

    def __enter__(self):
        self._gh_client = GithubClient(
            self.app_id,
            self.installation_id,
            self.private_key_file,
            self.repo_slug,
            self.user_agent_prefix,
        )
        return self._gh_client

    def __exit__(self, exception_type, exception_value, traceback):
        del self._gh_client
        return any(v is None
                   for v in (exception_type, exception_value, traceback))


@cache_once()
def get_app_key(key_path):
    with open(key_path) as f:
        return f.read()


@cache_once()
def get_github_integration(app_id, key_path):
    private_key = get_app_key(key_path)
    return github.GithubIntegration(app_id, private_key)


def get_installation_auth_token(gh_integration, install_id):
    return gh_integration.get_access_token(install_id).token


def get_installation_client(gh_integration, install_id):
    return github.Github(get_installation_auth_token(gh_integration, install_id))



#! /usr/bin/env python

import json
import sys
from warnings import catch_warnings

import click
from envparse import Env

from .github_api import GithubAPI


BASE_ENV_FILE_NAME = 'gh_chk_in_vars'

with catch_warnings(record=True):
    Env.read_envfile(f'/etc/{BASE_ENV_FILE_NAME}')
    Env.read_envfile(f'~/.config/.{BASE_ENV_FILE_NAME}')
    Env.read_envfile(f'~/.{BASE_ENV_FILE_NAME}')
    Env.read_envfile(f'.{BASE_ENV_FILE_NAME}')
    Env.read_envfile()


ENV_VAR_PREFIX = 'GH_CHK_IN_'
ENV_VAR_TMPL = f'{ENV_VAR_PREFIX}{{var_name}}'


@click.group()
@click.option('--private-key-file', prompt=True, envvar=ENV_VAR_TMPL.format(var_name='PRIVATE_KEY_FILE'))
@click.option('--app-id', prompt=True, envvar=ENV_VAR_TMPL.format(var_name='APP_ID'), type=int)
@click.option('--installation-id', prompt=True, envvar=ENV_VAR_TMPL.format(var_name='INSTALLATION_ID'), type=int)
@click.option('--repo-slug', prompt=True, envvar=ENV_VAR_TMPL.format(var_name='REPO_SLUG'))
@click.option('--user-agent', default=None)
@click.option('--name', prompt=True)
@click.option('--details-url')
@click.option('--external-id')
@click.option('--status')
@click.option('--started-at')
@click.option('--conclusion')
@click.option('--completed-at')
@click.option('--output')
@click.option('--actions')
@click.pass_context
def cli(ctx, **kwargs):
    app_id = kwargs.pop('app_id')
    installation_id = kwargs.pop('installation_id')
    private_key_file = kwargs.pop('private_key_file')
    repo_slug = kwargs.pop('repo_slug')
    user_agent = kwargs.pop('user_agent')
    gh_api = GithubAPI(app_id, installation_id, private_key_file, repo_slug, user_agent)
    ctx.obj.update(kwargs)
    ctx.obj['github_api'] = gh_api
    for opt in 'output', 'actions':
        try:
            ctx.obj[opt] = json.loads(ctx.obj[opt])
        except (KeyError, TypeError):
            """The option hasn't been provided."""


@cli.command()
@click.option('--head-branch', default='master')
@click.option('--head-sha', default='HEAD')
@click.pass_context
def post_check(ctx, head_branch, head_sha):
    try:
        with ctx.obj['github_api'] as gh_client:
            ctx.obj.pop('github_api')
            res = gh_client.post_check(head_branch, head_sha, ctx.obj)
        click.echo(f"Check Suite ID: {res['check_suite']['id']}")
        click.echo(f"Check Run ID: {res['id']}")
    except ValueError as val_exc:
        click.echo(val_exc)
        sys.exit(2)


@cli.command()
@click.option('--check-run-id', prompt=True, type=int)
@click.pass_context
def update_check(ctx, check_run_id):
    try:
        with ctx.obj['github_api'] as gh_client:
            ctx.obj.pop('github_api')
            res = gh_client.update_check(check_run_id, ctx.obj)
        click.echo(f"Check Suite ID: {res['check_suite']['id']}")
        click.echo(f"Check Run ID: {res['id']}")
    except ValueError as val_exc:
        click.echo(val_exc)
        sys.exit(2)


def main():
    return cli(obj={}, auto_envvar_prefix=ENV_VAR_PREFIX)


__name__ == '__main__' and main()



#! /usr/bin/env python
"""Distribution setup module."""

__requires__ = ('setuptools >=36.3.0', )
"""The list of pre-requisite requirements, needed to run this module."""

from setuptools import setup

setup_params = {
    'use_scm_version': True,
    'setup_requires': [
        'setuptools_scm>=1.15',
    ]
}

__name__ == '__main__' and setup(**setup_params)
